# encoding: utf-8

"""MIME-encoded electronic mail message classes."""

from turbomail import release

import turbogears, re, os, email

import email.Message
from email import Encoders, Charset
from email.Message import Message as MIMEMessage
from email.Utils import formatdate
from email.MIMEMultipart import MIMEMultipart
from email.MIMEBase import MIMEBase
from email.MIMEText import MIMEText
from email.Header import Header


__all__ = ['Message', 'KIDMessage']
_rich_to_plain = re.compile(r"(<[^>]+>)")


class Message(object):
	"""Simple e-mail message class.
	
	Message provides a means to easily create e-mail messages to be
	sent through the Dispatch mechanism or MailPool.  Message provides
	various helper functions to correctly format plain text, dual plain
	text and rich text MIME encoded messages, as well as handle
	embedded and external attachments.
	
	All properties can be set from the constructor.
	
	Example usage::
	
		import turbomail
		message = turbomail.Message(
				"from@host.com",
				"to@host.com",
				"Subject",
				plain="This is a plain message."
			)
	
	E-mail addresses can be represented as any of the following:
	 - A string.
	 - A 2-tuple of ("Full Name", "name@host.tld")
	
	Encoding can be overridden on a per-message basis, but note that
	'utf-8-qp' modifies the default 'utf-8' behaviour to output
	quoted-printable, and you will have to change it back yourself if
	you want base64 encoding.
		
	@ivar _processed: Has the MIME-encoded message been generated?
	@type _processed: bool
	@ivar _dirty: Has there been changes since the MIME message was last
	              generated?
	@type _dirty: bool
	@ivar date: The Date header.  Must be correctly formatted.
	@type date: string
	@ivar recipient: The To header.  A string, 2-tuple, or list of
	                 strings or 2-tuples.
	@ivar sender: The From header.  A string or 2-tuple.
	@ivar organization: The Organization header.  I{Optional.}
	@type organization: string
	@ivar replyto: The X-Reply-To header.  A string or 2-tuple.
	               I{Optional.}
	@ivar disposition: The Disposition-Notification-To header.  A string
	                   or 2-tuple.  I{Optional.}
	@ivar cc: The CC header.  As per the recipient property.
	          I{Optional.}
	@ivar bcc: The BCC header.  As per the recipient property.
	           I{Optional.}
	@ivar encoding: Content encoding.  Pulled from I{mail.encoding},
	                defaults to 'us-ascii'.
	@type encoding: string
	@ivar priority: The X-Priority header, a number ranging from 1-5.
	                I{Optional.}  Default: B{3}
	@type priority: int
	@ivar subject: The Subject header.
	@type subject: string
	@ivar plain: The plain text content of the message.
	@type plain: string
	@ivar rich: The rich text (HTML) content of the message.  Plain text
	            content B{must} be available as well.
	@type rich: string
	@ivar attachments: A list of MIME-encoded attachments.
	@type attachments: list
	@ivar embedded: A list of MIME-encoded embedded obejects for use in
	                the text/html part.
	@type embedded: list
	@ivar headers: A list of additional headers.  Can be added in a wide
	               variety of formats: a list of strings, list of
	               tuples, a dictionary, etc.  Look at the code.
	@ivar smtpfrom: The envelope address, if different than the sender.
	"""
	
	def __init__(self, sender=None, recipient=None, subject=None, **kw):
		"""Instantiate a new Message object.
		
		No arguments are required, as everything can be set using class
		properties.  Alternatively, I{everything} can be set using the
		constructor, using named arguments.  The first three positional
		arguments can be used to quickly prepare a simple message.
		
		An instance of Message is callable.
		
		@param sender: The e-mail address of the sender.  This is
					   encoded as the "From:" SMTP header.
		@type sender: string
		
		@param recipient: The recipient of the message.	 This gets
						  encoded as the "To:" SMTP header.
		@type recipient: string
		
		@param subject: The subject of the message.	 This gets encoded
						as the "Subject:" SMTP header.
		@type subject: string
		"""
		
		super(Message, self).__init__()
		
		self._processed = False
		self._dirty = False
		
		self.date = formatdate(localtime=True)
		self.recipient = recipient
		self.sender = sender
		self.organization = None
		self.replyto = None
		self.disposition = None
		self.cc = []
		self.bcc = []
		self.encoding = turbogears.config.get("mail.encoding", 'us-ascii')
		self.priority = 3
		self.subject = subject
		self.plain = None
		self.rich = None
		self.attachments = []
		self.embedded = []
		self.headers = []
		self.smtpfrom = None
		
		for i, j in kw.iteritems():
			if i in self.__dict__:
				self.__dict__[i] = j
			
	def attach(self, file, name=None):
		"""Attach an on-disk file to this message.
		
		@param file: The path to the file you wish to attach.
		@type file: string
		
		@param name: You can optionally override the filename of the
		             attached file.  This name will appear in the
		             recipient's mail viewer.  B{Optional.}
		@type name: string
		"""
		
		part = MIMEBase('application', "octet-stream")

		fp = open(file, "rb")
		part.set_payload(fp.read())
		Encoders.encode_base64(part)

		part.add_header('Content-Disposition', 'attachment', filename=os.path.basename([name, file][name is None]))
			
		self.attachments.append(part)
	
	def embed(self, file, name):
		"""Attach an on-disk image file and prepare for HTML embedding.
		
		This method should only be used to embed images.
		
		@param file: The path to the file you wish to attach.
		@type file: string
		
		@param name: The CID name to use for the embedded object.
		@type name: string
		"""
		
		from email.MIMEImage import MIMEImage
		
		fp = open(file, 'rb')
		part = MIMEImage(fp.read())
		fp.close()
		
		part.add_header('Content-ID', '<%s>' % name)
		
		self.embedded.append(part)
	
	def _normalize(self, addresslist):
		"""A utility function to return a list of addresses as a string."""
		
		addresses = []
		for i in [[addresslist], addresslist][type(addresslist) == type([])]:
			if type(i) == type(()):
				addresses.append('"%s" <%s>' % (i[0], i[1]))
			else: addresses.append(i)
		
		return ",\n ".join(addresses)
	
	def _process(self):
		"""Produce the final MIME message.
		
		Additinoally, if only a rich text part exits, strip the HTML to
		produce the plain text part.  (This produces identical output as
		KID, although lacks reverse entity conversion -- &amp;, etc.)
		"""
		
		if self.encoding == 'utf-8-qp':
			Charset.add_charset('utf-8', Charset.SHORTEST, Charset.QP, 'utf-8')
			self.encoding = 'utf-8'
		
		if self.rich and not self.plain:
			self.plain = _rich_to_plain.sub('', self.rich)
		
		if not self.rich:
			if not self.attachments:
				message = MIMEText(self.plain.encode(self.encoding), 'plain', self.encoding)
				
			else:
				message = MIMEMultipart()
				message.attach(MIMEText(self.plain.encode(self.encoding), 'plain', self.encoding))
				
		else:
			if not self.attachments:
				message = MIMEMultipart('alternative')
				message.attach(MIMEText(self.plain.encode(self.encoding), 'plain', self.encoding))
				
				if not self.embedded:
					message.attach(MIMEText(self.rich.encode(self.encoding), 'html', self.encoding))
				else:
					related = MIMEMultipart('related')
					message.attach(related)
					related.attach(MIMEText(self.rich.encode(self.encoding), 'html', self.encoding))
					
					for attachment in self.embedded:
						related.attach(attachment)
			
			else:
				message = MIMEMultipart()
				alternative = MIMEMultipart('alternative')
				message.attach(alternative)

				alternative.attach(MIMEText(self.plain.encode(self.encoding), 'plain', self.encoding))
				
				if not self.embedded:
					alternative.attach(MIMEText(self.rich.encode(self.encoding), 'html', self.encoding))
				else:
					related = MIMEMultipart('related')
					alternative.attach(related)
					related.attach(MIMEText(self.rich.encode(self.encoding), 'html', self.encoding))
					
					for attachment in self.embedded:
						related.attach(attachment)
			
		for attachment in self.attachments:
			message.attach(attachment)

		message.add_header('From', self._normalize(self.sender))
		message.add_header('Subject', self.subject)
		message.add_header('Date', formatdate(localtime=True))
		message.add_header('To', self._normalize(self.recipient))
		if self.replyto: message.add_header('Reply-To', self._normalize(self.replyto))
		if self.cc: message.add_header('Cc', self._normalize(self.cc))
		if self.disposition: message.add_header('Disposition-Notification-To', self._normalize(self.disposition))
		if self.organization: message.add_header('Organization', self.organization)
		if self.priority != 3: message.add_header('X-Priority', self.priority)

		if not self.smtpfrom:
			if type(self.sender) == type([]) and len(self.sender) > 1:
				message.add_header('Sender', self._normalize(self.sender[0]))
				message.add_header('Return-Path', self._normalize(self.sender[0]))
			else:
				message.add_header('Return-Path', self._normalize(self.sender))
		else:
			message.add_header('Return-Path', self._normalize(self.sender))
			message.add_header('Sender', self._normalize(self.smtpfrom))
			message.add_header('Return-Path', self._normalize(self.smtpfrom))
			message.add_header('Old-Return-Path', self._normalize(self.smtpfrom))

		message.add_header('X-Mailer', "TurboMail TurboGears Extension v.%s" % release.version)
		
		if type(self.headers) == type(()):
			for header in self.headers:
				if type(header) in [type(()), type([])]:
					message.add_header(*header)
				elif type(header) == type({}):
					message.add_header(**header)
		
		if type(self.headers) == type({}):
			for name, header in self.headers.iteritems():
				if type(header) in [type(()), type([])]:
					message.add_header(name, *header)
				elif type(header) == type({}):
					message.add_header(name, **header)
				else:
					message.add_header(name, header)
		
		self._message = message
		self._processed = True
		self._dirty = False
	
	def __setattr__(self, name, value):
		"""Set the dirty flag as properties are updated."""
		
		self.__dict__[name] = value
		if name != '_dirty': self.__dict__['_dirty'] = True
	
	def __call__(self):
		"""Produce a valid MIME-encoded message and return valid input
		for the Dispatch class to process.

		@return: Returns a tuple containing sender and recipient e-mail
		         addresses and the string output of MIMEMultipart.
		@rtype: tuple
		"""
		
		if not self._processed or self._dirty:
			self._process()
		
		recipients = []
		
		if isinstance(self.recipient, list):
			recipients.extend(self.recipient)
		else: recipients.append(self.recipient)
		
		if isinstance(self.cc, list):
			recipients.extend(self.cc)
		else: recipients.append(self.cc)
		
		if isinstance(self.bcc, list):
			recipients.extend(self.bcc)
		else: recipients.append(self.bcc)
		
		return dict(
				sender=self.sender,
				to=[[self.recipient], self.recipient][isinstance(self.recipient, list)],
				recipients=[i[1] for i in recipients if isinstance(i, tuple)] + [i for i in recipients if not isinstance(i, tuple)],
				subject=self.subject,
				message=self._message.as_string(),
			)


class KIDMessage(Message):
	"""A template which accepts a named template with arguments.
	
	Example usage::
	
		import turbomail
		message = turbomail.KIDMessage(
				"from@host.com",
				"to@host.com",
				"Subject",
				"app.templates.mail",
				dict()
			)
	
	Do not specify message.plain or message.rich content - the template
	will override what you set.  If you wish to hand-produce content,
	use the Message class.
	"""
	
	def __init__(self, sender, recipient, subject, template, variables={}, **kw):
		"""Store the additonal template and variable information.
		
		@param template: A dot-path to a valid KID template.
		@type template: string
		
		@param variables: A dictionary containing named variables to
		                  pass to the template engine.
		@type variables: dict
		"""
		
		self._template = template
		self._variables = dict(sender=sender, recipient=recipient, subject=subject)
		self._variables.update(variables)
				
		super(KIDMessage, self).__init__(sender, recipient, subject, **kw)
	
	def _process(self):
		"""Automatically generate the plain and rich text content."""
		
		turbogears.view.base._load_engines()
		
		data = dict()
		
		for (i, j) in self._variables.iteritems():
			if callable(j): data[i] = j()
			else: data[i] = j
		
		self.plain = turbogears.view.engines.get('kid').render(data, format="plain", template=self._template)
		self.rich = turbogears.view.engines.get('kid').render(data, template=self._template)
		
		return super(KIDMessage, self)._process()



# encoding: utf-8

"""
Introduction
============

	TurboMail is a TurboGears extension - meaning that it starts up and
	shuts down alongside TurboGears applications you write in the same
	way that visit tracking and identity do.  TurboMail uses built-in
	Python modules for SMTP communication and MIME e-mail creation, but
	greatly simplifies these tasks by performing the grunt-work for you.
	
	Being multi-threaded, TurboMail allows you to enqueue messages to be
	sent and then immediately continue with processing, resulting in a
	much more fluid user experience.  Threads are handled intelligently
	(increasing the number of threads as demand increases) and they are
	automatically recycled.  There is only ever one SMTP connection per
	thread.
	
	Benchmarking
	------------
	
		Throughput using the default options is sufficient for most use:
		100 messages in 45 seconds; just over 2 messages a second. Using
		a greater number of threads, 10 vs. 4, 100 messags take 30
		seconds; just over 3 messages a second.  YMMV.  Note that if a
		thread is idle, it will immediately deliver requests added to
		the queue, thus increasing the idle time will increase sparse
		performance.
	
	TurboMail is heavily inspired by PHPMailer, a very, very handy class
	for PHP 4 & 5 by Brent R. Matzelle.
	

Installation
============

	Simply easy_install the package::

		easy_install TurboMail

	TurboMail installs no external scripts.

Upgrade
=======

	Upgrading also uses easy_install::
	
		easy_install -U TurboMail

Configuration
=============

	TurboMail understands a large number of configuration options, all
	piggy-backed from your application's configuration.  Organized into
	two groups, the advanced set can be safely ignored in most
	applications.  Each option is listed with its default value.
	
	Simple Options
	--------------
	
		 - I{mail.on} (Default: B{False}) Enable TurboMail.  B{Required.}
		 - I{mail.server} (Default: B{None}) SMTP server address.
		   B{Required.}
		 - I{mail.username} (Default: B{None})
		 - I{mail.password} (Default: B{None})
	
		Both a username and password are required to enable
		authentication.
	
	Advanced Options
	----------------
	
		 - I{mail.debug} (Default: B{False}) Output all SMTP server
		   communications.
		 - I{mail.interval} (Default: B{10}) Polling delay between new
		   thread creation, in seconds.
		 - I{mail.threads} (Default: B{4}) Maximum number of concurrent
		   threads.
		 - I{mail.jobs} (Default: B{10}) Maximum number of job units per
		   thread.
		 - I{mail.timeout} (Default: B{60}) Maximum time a worker thread
		   will wait for additional jobs, in seconds.
		 - I{mail.tls} (Defalut: None) Enable or disable TLS, None will
		   attempt to auto-detect TLS.  This will not always work.
		 - I{mail.encoding} (Default: B{'us-ascii'}) Set the character
		   set and encoding on the MIMEText parts of the message.
		   Common character sets include:
		    - us-ascii - I{Performs no encoding, but is 7bit only.}
		    - iso-8859-1 - I{Uses quoted-printable encoding.}
		    - utf-8 - I{Uses base64 encoding.}
		   Due to the way Python's email package handles character sets,
		   the following additional virtual character sets are provided
		   by TurboMail, and will override the global defaults:
		    - utf-8-qp - I{Sets utf-8 to use quoted-printable encoding.}
		   Headers are not encoded.  DIY.
		 - I{mail.polling} (Defaut: B{False}) If enabled, configures the
		   thread pool to poll every I{mail.interval} seconds for new
		   jobs.  This may give performance benefits to the running
		   application.  The default behaviour is to create new threads
		   as soon as work is enqueued, resulting in faster delivery.
		
		In debug mode using a single thread with a maximum of one job
		can be advantageous.  Having a single thread with a maximum of a
		single job limits TurboMail to a single SMTP connection at a
		time and automatically disconnects for I{each message}.

Basic Usage
===========

	To use TurboMail in your TurboGears application, after adding the
	appropriate configuration options to your application, perform the
	following steps:
	
	 1. Import TurboMail::
	
	    	import turbomail
	
	 2. Create a L{Message} or L{KIDMessage} object::
	
	    	message = turbomail.Message(from, to, subject)
	
	 3. Set some content for your message::
	
	    	message.plain = "Hello world!"
	
	 4. Enqueue your message::
	
	    	turbomail.enqueue(message)
	
	Your message will now have been enqueued.  It will take at most
	I{mail.interval} seconds for a thread to be created to deliver it.
	The best case scenario is if there is an idle thread waiting, in
	which case delivery will be immediate.

Advanced Usage - Logging
========================

	Additionally, you can configure your application to log TurboMail
	events differently, in a more loggable and machine parsable way.
	You do so by adding the following lines to the formatters section
	of your log.cfg::
	
		[[[timed_message]]]
		format='*(asctime)s *(message)s'
	
	Add the following to the handlers section of log.cfg::
	
		[[[mail_out]]]
		class='StreamHandler'
		level='INFO'
		args='(sys.stdout,)'
		formatter='timed_message'
	
	And finally, add the following to your dev.cfg::
	
		[[[mail]]]
		level='INFO'
		qualname='turbomail.dispatch'
		handlers=['mail_out']
		propagate=0
	
	If you wish to log mail dispatch to a file, for example in your
	production configuration, use this instead of the above::
	
		[[handlers]]
		[[[mail_out]]]
		args="('mail.log',)"
		class='StreamHandler'
		level='INFO'
		formatter='timed_message'
		
		[[loggers]]
		[[[mail]]]
		level='INFO'
		qualname='turbomail.dispatch'
		handlers=['mail_out']
		propagate=0
	
	The format of turbomail.dispatch INFO log entries is::
	
		[user@]server size ("from_name") from_addr ("to_name") to_addr - subject
	
	Designed for easy parsing and tracking of statistics, the log
	format uses the following conventions:
	
	 - Entries between square brackets are optional and may be omitted.
	 - Entries between round brackets may be replaced with a dash if
	   unavailable.
	 - The size field is in bytes and represents the total size of the
	   MIME-encoded message including headers, after character set
	   conversion.
	 - When sending to multiple recipients, the to_addr field becomes
	   the number of recipients wrapped in round brackets.  E.g. "(3)"
	 - The subject field extends to the EOL - quotes, dashes, and other
	   symbols should be treated as part of the subject.

Changelog
=========

	Version 1.0
	-----------
		 - Initial release.
	
	Version 1.0.1
	-------------
		 - Minor updates to remove unneeded arguments.
		 - Complete source-level epydoc documentation.
	
	Version 1.0.4.1
	---------------
		 - Better auto-detection of TLS capability.
		 - A new configuration directive, mail.tls; True, False, or
		   None to auto-detect.
		 - Fixes a bug in KIDMessage which rendered it
		   non-functional.
		 - Changed the behavior of a worker dying from old age to spawn
		   a new process immediately.
		 - Minor fixes and updates to the documentation.
		 - Benchmark results in the documentation.
	
	Version 1.0.4.2
	---------------
		 - Added encoding configuration directive.
		 - Encoding can be passed to a Message constructor to override
		   the encoding on a message-by-message basis.
	
	Version 1.1
	-----------
		 - Cleaned up the log output.
		 - Added documentation for logging.
		 - KID i18n session bug fixed in TurboGears 1.0 and trunk SVN.
		 - Marked as stable for the Python Cheese Shop.
	
	Version 2.0
	-----------
		 - Default thread creation mechanism is on-demand, not polling.
		   You can change back to the (old) polling mechanism by
		   setting the following configuration option::
		   	mail.polling = True
		 - TemplateMessage has been renamed KIDMessage.
		 - Can now use 'utf-8-qp' to configure the 'utf-8' charset for
		   quoted-printable encoding.
		 - MIME-encoded message generation was re-written.  Now, simple
		   plain-text-only messages have almost zero overhead.
		   Complexity of the generated document increases with feature
		   use.
		 - It is now safe to import enqueue from TurboMail - it no
		   longer polymorphs after TurboGears start-up and shutdown.
		 - Enhanced logging output - see above.
		 - Better tracking of when to rebuild the MIME message by using
		   a dirty flag.
		 - Many, many additional headers.  Look at the documentation for
		   the message class for more information.
		 - Multiple recipients.
		
		There is, however, an outstanding bug in KIDMessage.  When
		generating the plain-text alternative KID seems to default
		to ascii encoding, which bombs out if you use any extended
		characters in the template, or variables passed to the template.
		
		Patches are welcome!

@var _queue: After TurboGears startup within an application which has
            enabled TurboMail, I{queue} is an instance of a MailPool
            object.

"""

from turbomail.release import \
		version as __version__, \
		author as __author__, \
		email as __email__, \
		license as __license__, \
		copyright as __copyright__

import logging
log = logging.getLogger("turbomail")

from turbomail.startup import start_extension, shutdown_extension
from turbomail.exceptions import *
from turbomail.message import Message, KIDMessage


__all__ = ['MailException', 'MailNotEnabledException', 'MailConfigurationException', 'Message', 'KIDMessage', 'enqueue', 'dispatch', 'exceptions', 'message', 'pool', 'startup']


_queue = None

def enqueue(work):
	"""Enqueue a message in the message thread-pool queue."""
	
	if _queue is None:
		raise MailNotEnabledException
	
	else:
		_queue.enqueue(work)
